import base64
import itertools
import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Dict, Iterable, List

import tomlkit

from feeluown.consts import COLLECTIONS_DIR
from feeluown.utils.dispatch import Signal
from feeluown.library import resolve, reverse, ResolverNotFound, \
    ResolveFailed, ModelState, CollectionType
from feeluown.utils.utils import elfhash

logger = logging.getLogger(__name__)

COLL_LIBRARY_IDENTIFIER = 'library'
COLL_POOL_IDENTIFIER = 'pool'
# for backward compat, we should never change these filenames
LIBRARY_FILENAME = f'{COLL_LIBRARY_IDENTIFIER}.fuo'
POOL_FILENAME = f'{COLL_POOL_IDENTIFIER}.fuo'
DEPRECATED_FUO_FILENAMES = (
    'Songs.fuo', 'Albums.fuo', 'Artists.fuo', 'Videos.fuo'
)

TOML_DELIMLF = "+++\n"


class CollectionAlreadyExists(Exception):
    pass


class Collection:
    """
    TODO: This collection should be moved into local provider.
    """

    def __init__(self, fpath):
        # TODO: 以后考虑添加 identifier 字段，identifier
        # 字段应该尽量设计成可以跨电脑使用
        self.fpath = str(fpath)
        # TODO: 目前还没想好 collection identifier 计算方法，故添加这个函数
        # 现在把 fpath 当作 identifier 使用，但对外透明
        self.identifier = elfhash(base64.b64encode(bytes(self.fpath, 'utf-8')))

        # these variables should be inited during loading
        self.type = None
        self.name = None  # Collection title.
        self.models = []
        self.updated_at = None
        self.created_at = None
        self.description = None
        self._has_nonexistent_models = False

        #: tomkit.toml_document.Document
        self._metadata = None

    def load(self):
        """解析文件，初始化自己"""
        # pylint: disable=too-many-branches
        self.models = []
        filepath = Path(self.fpath)
        name = filepath.stem
        stat_result = filepath.stat()
        self.updated_at = datetime.fromtimestamp(stat_result.st_mtime)
        self.name = name
        if name == COLL_LIBRARY_IDENTIFIER:
            self.type = CollectionType.sys_library
        elif name == COLL_POOL_IDENTIFIER:
            self.type = CollectionType.sys_pool
        else:
            self.type = CollectionType.mixed

        # parse file content
        with filepath.open(encoding='utf-8') as f:
            first = f.readline()
            lines = []
            if first == TOML_DELIMLF:
                is_valid = False
                for line in f:
                    if line == TOML_DELIMLF:
                        is_valid = True
                        break
                    lines.append(line)
                if is_valid is True:
                    toml_str = ''.join(lines)
                    metadata = tomlkit.parse(toml_str)
                    self._loads_metadata(metadata)
                    lines = []
                else:
                    logger.warning('the metadata is invalid, will ignore it')
            else:
                lines.append(first)

            for line in itertools.chain(lines, f):
                if not line.strip():  # ignore empty lines
                    continue
                try:
                    model = resolve(line)
                except ResolverNotFound:
                    logger.warning('resolver not found for line:%s', line)
                    model = None
                except ResolveFailed as e:
                    logger.warning('resolve failed, file:%s, line:%s, error:%s',
                                   str(filepath), line, str(e))
                    model = None
                if model is not None:
                    if model.state is ModelState.not_exists:
                        self._has_nonexistent_models = True
                    self.models.append(model)

    @classmethod
    def create_empty(cls, fpath, title=''):
        """Create an empty collection."""
        if os.path.exists(fpath):
            raise CollectionAlreadyExists()

        doc = tomlkit.document()
        if title:
            doc.add('title', title)         # type: ignore[arg-type]
        doc.add('created', datetime.now())  # type: ignore[arg-type]
        doc.add('updated', datetime.now())  # type: ignore[arg-type]
        with open(fpath, 'w', encoding='utf-8') as f:
            f.write(TOML_DELIMLF)
            f.write(tomlkit.dumps(doc))
            f.write(TOML_DELIMLF)

        coll = cls(fpath)
        coll._loads_metadata(doc)
        coll.type = CollectionType.mixed
        return coll

    def add(self, model):
        """add model to collection

        :param model: :class:`feeluown.library.BaseModel`
        :return: True means succeed, False means failed
        """
        if model not in self.models:
            line = reverse(model, as_line=True)
            with open(self.fpath, 'r+', encoding='utf-8') as f:
                content = f.read()
                parts = content.split(TOML_DELIMLF, maxsplit=2)
                body = parts[-1]
                f.seek(0, 0)
                self._write_metadata_if_needed(f)
                f.write(f'{line}\n{body}')
                f.truncate()
            self.models.insert(0, model)
        return True

    def remove(self, model):
        if model in self.models:
            url = reverse(model)
            with open(self.fpath, 'r+', encoding='utf-8') as f:
                content = f.read()
                parts = content.split(TOML_DELIMLF, maxsplit=2)
                body = parts[-1]
                lines = []
                for line in body.split('\n'):
                    if line.startswith(url):
                        continue
                    if line:
                        lines.append(line)

                f.seek(0)
                self._write_metadata_if_needed(f)
                f.write('\n'.join(lines))
                f.truncate()
                # 确保最后写入一个换行符，让文件更加美观
                if lines and not lines[-1].endswith('\n'):
                    f.write('\n')
            self.models.remove(model)
        return True

    def on_provider_added(self, provider):
        if not self._has_nonexistent_models:
            return
        for i, model in enumerate(self.models.copy()):
            if model.state is ModelState.not_exists and \
                    model.source == provider.identifier:
                new_model = resolve(reverse(model, as_line=True))
                # TODO: emit data changed signal
                self.models[i] = new_model
        # TODO: set _has_nonexistent_models to proper value

    def on_provider_removed(self, provider):
        for model in self.models:
            if model.source == provider.identifier:
                model.state = ModelState.not_exists
                self._has_nonexistent_models = True

    def _loads_metadata(self, metadata):
        self._metadata = metadata
        self.created_at = metadata.get('created')
        self.updated_at = metadata.get('updated', self.updated_at)
        self.name = metadata.get('title', self.name)
        self.description = metadata.get('description')

    def _write_metadata_if_needed(self, f):
        # write metadata only if it had before
        if self._metadata:
            self.updated_at = self._metadata['updated'] = datetime.now()
            s = tomlkit.dumps(self._metadata)
            f.write(TOML_DELIMLF)
            f.write(s)
            if not s.endswith('\n'):
                f.write('\n')
            f.write(TOML_DELIMLF)


class CollectionManager:

    def __init__(self, app):
        self._app = app
        self.scan_finished = Signal()
        self._library = app.library
        self.default_dir = COLLECTIONS_DIR

        self._id_coll_mapping: Dict[int, Collection] = {}
        self._sys_colls = {}

    def get(self, identifier):
        if identifier in (CollectionType.sys_pool, CollectionType.sys_library):
            return self._sys_colls[identifier]
        return self._id_coll_mapping[int(identifier)]

    def get_coll_library(self):
        for coll in self._id_coll_mapping.values():
            if coll.type == CollectionType.sys_library:
                return coll
        assert False, "collection 'library' must exists."

    def create(self, fname, title) -> Collection:
        first_valid_dir = ''
        for d, exists in self._get_dirs():
            if exists:
                first_valid_dir = d
                break

        assert first_valid_dir, 'there must be a valid collection dir'
        normalized_name = fname.replace(' ', '_')
        fpath = os.path.join(first_valid_dir, normalized_name)
        filepath = f'{fpath}.fuo'
        logger.info(f'Create collection:{title} at {filepath}')
        return Collection.create_empty(filepath, title)

    def remove(self, collection: Collection):
        coll_id = collection.identifier
        if coll_id in self._id_coll_mapping:
            self._id_coll_mapping.pop(coll_id)
            os.remove(collection.fpath)

    def _get_dirs(self, ):
        directorys = [self.default_dir]
        if self._app.config.COLLECTIONS_DIR:
            if isinstance(self._app.config.COLLECTIONS_DIR, list):
                directorys += self._app.config.COLLECTIONS_DIR
            else:
                directorys.append(self._app.config.COLLECTIONS_DIR)
        expanded_dirs = []
        for directory in directorys:
            directory = os.path.expanduser(directory)
            expanded_dirs.append((directory, os.path.exists(directory)))
        return expanded_dirs

    def scan(self):
        colls: List[Collection] = []
        for coll in self._scan():
            if coll.type == CollectionType.sys_library:
                self._sys_colls[CollectionType.sys_library] = coll
            elif coll.type == CollectionType.sys_pool:
                self._sys_colls[CollectionType.sys_pool] = coll
            else:
                colls.append(coll)

        if CollectionType.sys_pool not in self._sys_colls:
            pool_fpath = os.path.join(self.default_dir, POOL_FILENAME)
            assert not os.path.exists(pool_fpath)
            logger.info('Generating collection pool.')
            coll = Collection.create_empty(pool_fpath, '想听')
            self._sys_colls[CollectionType.sys_pool] = coll

        pool_coll = self._sys_colls[CollectionType.sys_pool]
        library_coll = self._sys_colls[CollectionType.sys_library]
        colls.insert(0, pool_coll)
        colls.insert(0, library_coll)
        for collection in colls:
            coll_id = collection.identifier
            assert coll_id not in self._id_coll_mapping, collection.fpath
            self._id_coll_mapping[coll_id] = collection
        self.scan_finished.emit()

    def refresh(self):
        self.clear()
        self.scan()

    def listall(self):
        return self._id_coll_mapping.values()

    def clear(self):
        self._id_coll_mapping.clear()

    def _scan(self) -> Iterable[Collection]:
        """Scan collections directories for valid fuo files, yield
        Collection instance for each file.
        """
        default_fpaths = []
        valid_dirs = self._get_dirs()
        for directory, exists in valid_dirs:
            if not exists:
                logger.warning('Collection directory %s does not exist', directory)
                continue
            for filename in os.listdir(directory):
                if not filename.endswith('.fuo'):
                    continue
                filepath = os.path.join(directory, filename)
                if filename in DEPRECATED_FUO_FILENAMES:
                    default_fpaths.append(filepath)
                    continue
                coll = Collection(filepath)
                coll.load()
                self._app.library.provider_added.connect(coll.on_provider_added)
                self._app.library.provider_removed.connect(coll.on_provider_removed)
                yield coll

        fpath, generated = self.generate_library_coll_if_needed(default_fpaths)
        # Avoid to yield a duplicated collection.
        if generated is True:
            coll = Collection(fpath)
            coll.load()
            self._app.library.provider_added.connect(coll.on_provider_added)
            self._app.library.provider_removed.connect(coll.on_provider_removed)
            yield coll

    def generate_library_coll_if_needed(self, default_fpaths):
        """
        :return: (library file path, generated)
        """
        library_fpath = os.path.join(self.default_dir, LIBRARY_FILENAME)
        if os.path.exists(library_fpath):
            if default_fpaths:
                paths_str = ','.join(default_fpaths)
                logger.warning(f'{paths_str} are ignored since {library_fpath} exists')
            return library_fpath, False

        logger.info('Generating collection library')
        lines = [TOML_DELIMLF,
                 'title = "Library"',
                 TOML_DELIMLF]
        if default_fpaths:
            for fpath in default_fpaths:
                with open(fpath, 'r', encoding='utf-8') as f:
                    for line in f:
                        if line.startswith('fuo://'):
                            lines.append(line)
        with open(library_fpath, 'w', encoding='utf-8') as f:
            f.write('\n'.join(lines))

        # backup old sys collections if they exists
        for fpath in default_fpaths:
            dirname = os.path.dirname(fpath)
            filename = os.path.basename(fpath)
            filename += '.bak'
            new_fpath = os.path.join(dirname, filename)
            logger.info(f'Rename {fpath} to {new_fpath}')
            os.rename(fpath, new_fpath)
        return library_fpath, True



from PyQt5.QtCore import Qt, QRectF, QPoint
from PyQt5.QtGui import QTextOption
from PyQt5.QtWidgets import QToolTip

from feeluown.gui.widgets.statusline import StatuslineLabel


class NotifyStatus(StatuslineLabel):
    def __init__(self, app, parent=None):
        super().__init__(app, parent)
        self._app = app

    def drawInner(self, painter):
        inner_rect = QRectF(0, 0, self._inner_width, self._inner_height)
        painter.drawText(inner_rect, '✉', QTextOption(Qt.AlignCenter))

    def show_msg(self, text, timeout=1500):
        QToolTip.showText(self.mapToGlobal(QPoint(0, 0)),
                          text,
                          self,
                          self.rect(),
                          timeout)



import sys

if sys.platform == 'win32':
    # By default, it uses SimSun(宋体) on windows, which is a little ugly.
    # "Segoe UI Symbol" is used to render charactor symbols.
    # "Microsoft Yahei" is used to render chinese (and english).
    # Choose a default sans-serif font when the first two fonts do not work,
    FontFamilies = ['Segoe UI Symbol', 'Microsoft YaHei', 'sans-serif']
elif sys.platform == 'darwin':
    FontFamilies = ['arial', 'Hiragino Sans GB', 'sans-serif']
else:
    FontFamilies = [
        'Helvetica Neue', 'Helvetica', 'Arial', 'Microsoft Yahei', 'Hiragino Sans GB',
        'Heiti SC', 'WenQuanYi Micro Hei', 'sans-serif'
    ]

# The width of ScrollBar on macOS is about 10.
ScrollBarWidth = 10



# -*- coding: utf-8 -*-

import logging
import sys
import socket
import time
from collections import OrderedDict
from copy import copy, deepcopy
from functools import wraps
from itertools import filterfalse

logger = logging.getLogger(__name__)


def win32_is_port_binded(host, port):
    """
    sock.connect_ex may block for 2 second if port is not used on Windows.
    This may be the fastest way to check if a port is bined on Windows.
    Remember that (127.0.0.1, port) and (0.0.0.0, port) are different for `sock.bind`.
    """
    import errno  # pylint: disable=import-outside-toplevel

    inuse = False
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.bind((host, port))
    except socket.error as e:
        if e.errno == errno.EADDRINUSE:
            inuse = True
        else:
            raise
    finally:
        sock.close()
    return inuse


def is_port_inuse(port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    rv = sock.connect_ex(('127.0.0.1', port))
    sock.close()
    return rv == 0


def parse_ms(ms):
    minute = int(ms / 60000)
    second = int((ms % 60000) / 1000)
    return minute, second


def log_exectime(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        t = time.process_time()
        result = func(*args, **kwargs)
        elapsed_time = time.process_time() - t
        logger.debug('function %s executed time: %f ms',
                     func.__name__, elapsed_time * 1000)
        return result
    return wrapper


def elfhash(s):
    """
    :param string: bytes

    >>> import base64
    >>> s = base64.b64encode(b'hello world')
    >>> elfhash(s)
    224648685
    """
    hash = 0
    x = 0
    for c in s:
        hash = (hash << 4) + c
        x = hash & 0xF0000000
        if x:
            hash ^= (x >> 24)
            hash &= ~x
    return (hash & 0x7FFFFFFF)


class DedupList(list):
    """List that doesn't contain duplicate items

    The item should properly implement __hash__ and __eq__. If items
    are not equal to each other, they must not have same hash.
    """

    def _get_index(self, index):
        """ project idx into range(len) """
        if index <= -len(self):
            return 0
        if index < 0:
            return index + len(self)
        if index >= len(self):
            return len(self)
        return index

    @staticmethod
    def dic():
        """ return a dict that remembers insertion order """
        return {} if sys.version_info[1] > 5 else OrderedDict()

    def __init__(self, seq=(), dedup=True):
        if dedup:
            seq = self.dic().fromkeys(seq)
        # keep idx in a dict for dedup and index
        self._map = dict(zip(seq, range(len(seq))))
        super().__init__(seq)

    def __getitem__(self, item):
        result = super().__getitem__(item)
        if isinstance(item, slice):
            # Always return a DedupList when slicing to avoid accidentally
            # convert a DedupList to normal list.
            return DedupList(result, dedup=False)
        else:
            return result

    def __add__(self, other):
        if isinstance(other, list):
            result = copy(self)
            result.extend(other)
            return result
        raise TypeError("can only concatenate list to DedupList")

    def __radd__(self, other):
        if isinstance(other, list):
            if isinstance(other, DedupList):
                result = copy(other)
            else:
                # To avoid accidentally convert a DedupList to normal list.
                result = DedupList(other)
            result.extend(self)
            return result
        raise TypeError("invalid concat")

    def __setitem__(self, key, value):
        if value in self._map:
            raise ValueError("item already exists in DedupList")
        self._map.pop(self[key])
        self._map[value] = key
        super().__setitem__(key, value)

    def __contains__(self, item):
        return item in self._map

    def __copy__(self):
        result = DedupList(self, dedup=False)
        return result

    def __deepcopy__(self, memo):
        inter_list = [deepcopy(item) for item in self]
        result = DedupList(inter_list, dedup=False)
        memo[id(self)] = result
        return result

    def swap(self, idx_1, idx_2):
        item_1 = self[idx_1]
        item_2 = self[idx_2]
        self._map[item_1] = idx_2
        self._map[item_2] = idx_1
        super().__setitem__(idx_1, item_2)
        super().__setitem__(idx_2, item_1)

    def sort(self, *args, **kwargs):
        super().sort(*args, **kwargs)
        self._map = dict(zip(self, range(len(self))))

    def append(self, obj):
        if obj not in self._map:
            self._map[obj] = len(self)
            super().append(obj)

    def extend(self, iterable):
        length = len(self)
        append_list = list(filterfalse(self._map.__contains__, iterable))   # dedup
        self._map.update(zip(append_list, range(length, length + len(append_list))))
        super().extend(append_list)

    def index(self, object, start: int = None, stop: int = None):
        if object not in self._map:
            raise ValueError("not in list")
        # get idx from _map directly
        idx = self._map[object]
        if start:
            if not stop:
                stop = len(self)
            if idx not in range(start, stop):
                raise ValueError("not in list")
        return idx

    def insert(self, index: int, obj):
        if obj not in self._map:
            # insert accepts out-of-range indices, but we don't want them in our _map
            index = self._get_index(index)
            # all idx in _map after 'index' should be changed
            for key, idx in self._map.items():
                if idx >= index:
                    self._map[key] = idx + 1
            self._map[obj] = index
            super().insert(index, obj)

    def pop(self, index: int = None):
        index = index if index is not None else -1
        item = super().pop(index)
        # list.pop() returns an index, so no need to calculate manually like 'insert'
        index = self._map.pop(item)
        # all idx in _map after 'index' should be changed
        for obj, idx in self._map.items():
            if idx > index:
                self._map[obj] = idx - 1
        return item

    def remove(self, item):
        if item not in self._map:
            # raise same exception as list
            raise ValueError("list.remove(x): x not in list")
        idx = self._map[item]
        self.pop(idx)

    def clear(self):
        self._map.clear()
        super().clear()


def int_to_human_readable(i):
    if i >= 100000000:
        return f'{i / 100000000:.1f}亿'
    elif i >= 10000:
        return f'{i / 10000:.1f}万'
    return i



# -*- coding: utf-8 -*-
from pydantic import ConfigDict, BaseModel, Field


DEFAULT_TITLE = DEFAULT_ARTIST_NAME = DEFAULT_ALBUM_NAME = 'Unknown'


class Common(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="ignore")

    duration: float
    title: str = DEFAULT_TITLE

    artists_name: str = Field(DEFAULT_ARTIST_NAME, alias='artist')
    album_name: str = Field(DEFAULT_ALBUM_NAME, alias='album')
    album_artist_name: str = Field(DEFAULT_ARTIST_NAME, alias='albumartist')

    track: str = Field('1/1', alias='tracknumber')
    disc: str = Field('1/1', alias='discnumber')

    date: str = ''
    genre: str = ''


class EasyMP3Model(Common):
    pass


class APEModel(Common):
    pass


class FLACModel(Common):
    track_number: int = Field(1, alias='tracknumber')
    disc_number: int = Field(1, alias='discnumber')
    track_total: int = Field(1, alias='tracktotal')
    disc_total: int = Field(1, alias='disctotal')



# flake8: noqa

import asyncio
import sys


def patch_qeventloop():
    # patch note::
    #
    # asyncio.wrap_future should set loop to self, otherwise, we can not call run_in_executor
    # in another thread(the loop not belongs to). The event loop implemented in asyncio
    # library also does this.

    from .compat import QEventLoop, QThreadExecutor

    def run_in_executor(self, executor, callback, *args):
        """Run callback in executor.

        If no executor is provided, the default executor will be used, which defers execution to
        a background thread.
        """
        # pylint: disable=all

        self._logger.debug('Running callback {} with args {} in executor'.format(callback, args))
        if isinstance(callback, asyncio.Handle):
            assert not args
            assert not isinstance(callback, asyncio.TimerHandle)
            if callback._cancelled:
                f = asyncio.Future()
                f.set_result(None)
                return f
            callback, args = callback.callback, callback.args

        if executor is None:
            self._logger.debug('Using default executor')
            executor = self.get_default_executor()

        if executor is None:
            self._logger.debug('Creating default executor')
            self.set_default_executor_v2(QThreadExecutor())
            executor = self.get_default_executor()

        return asyncio.wrap_future(executor.submit(callback, *args), loop=self)

    def get_default_executor(self):
        return getattr(self, 'default_executor', None)

    def set_default_executor_v2(self, executor):
        setattr(self, 'default_executor', executor)

        # QEventLoop will shutdown executor when loop is closed
        self.set_default_executor(executor)

    QEventLoop.get_default_executor = get_default_executor
    QEventLoop.set_default_executor_v2 = set_default_executor_v2
    QEventLoop.run_in_executor = run_in_executor


def patch_janus():
    """
    https://github.com/aio-libs/janus/issues/250
    """

    import janus

    janus.current_loop = asyncio.get_event_loop


def patch_mutagen():

    from mutagen.id3 import _specs

    old_decode_terminated = _specs.decode_terminated

    def decode_terminated(data, encoding, strict=True):
        """
        国内部分歌曲的 id3 信息是使用 GBK 编码，但是它的元信息显示自己是
        latin1 编码，所以 mutagen 解析的时候会返回不正确的结果。

        这里，我们没有选择重新实现原函数，而是对原函数的返回进行加工，
        这样可以比较好地确保我们的实现不会影响到 mutagen 的其它逻辑。
        但这样实现有一个问题，原函数已经进行过 decode 和 encode，我们这里又重复
        encode 和 decode，理论上会带来一定的性能损耗，欢迎有兴趣的同鞋进行优化。

        ps: 根据 id3 2.3/2.4 的规范，gbk 编码 *似乎* 本来就是非法的，
        规范上说只支持这四种： latin1, utf16, utf_16_be, utf8，
        mutagen 确实也是按照 spec 来实现的，但现实很残酷。
        """
        orig_value, data = old_decode_terminated(data, encoding, strict)
        value = orig_value
        if encoding == 'latin1':
            value = orig_value.encode('latin1')
            # 参考 deepin-music-player 项目 easymp3.py 实现
            for codec in ['gbk', 'big5', 'utf-8', 'iso-8559-15', 'cp1251']:
                try:
                    value = value.decode(codec)
                except (UnicodeDecodeError, LookupError):
                    # decode failed or codec not found
                    pass
                else:
                    break
            else:
                value = orig_value
        return value, data

    _specs.decode_terminated = decode_terminated


def patch_pydantic():
    """
    For some reason, some plugins already use pydantic v2 and they only use
    pydantic.v1 package. Make them compat with pydantic v1 in an hacker way :)
    """
    try:
        import pydantic.v1
    except ImportError:
        import pydantic
        sys.modules['pydantic.v1'] = pydantic



from PyQt5.QtCore import QPoint, Qt, QRect, QRectF, QTimer, QPointF
from PyQt5.QtWidgets import QPushButton, QStyle, QStyleOptionButton
from PyQt5.QtGui import QPainter, QPalette, QPainterPath

from feeluown.gui.drawers import (
    HomeIconDrawer,
    PlusIconDrawer,
    TriangleIconDrawer,
    CalendarIconDrawer,
    RankIconDrawer,
    StarIconDrawer,
    VolumeIconDrawer,
    SearchIconDrawer,
    FireIconDrawer,
    EmojiIconDrawer,
)
from feeluown.gui.helpers import darker_or_lighter, painter_save


def set_pen_width(painter, width):
    pen = painter.pen()
    pen.setWidthF(width)
    painter.setPen(pen)


def set_pen_1_5(painter):
    set_pen_width(painter, 1.5)


class SelfPaintAbstractButton(QPushButton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # It seems macOS platform does not turn this attribute on by default.
        # The following two attrs are only needed on macOS, and they don't
        # cause any side effects on other platforms.
        self.setAttribute(Qt.WA_Hover, True)
        self.setAttribute(Qt.WA_LayoutUsesWidgetRect, True)

    def paintEvent(self, _):
        raise NotImplementedError('paintEvent must be implemented')

    def hitButton(self, pos: QPoint) -> bool:
        # QPushButton use style().subElementRect().contains to check if it is hit.
        # style().subElementRect is really small for buttons. So it does not
        # work well for small buttons (tested on macOS).
        return self.rect().contains(pos)

    def paint_border_bg_when_hover(self, painter, radius=3):
        opt = QStyleOptionButton()
        self.initStyleOption(opt)

        if opt.state & QStyle.State_MouseOver:
            painter.save()
            painter.setPen(Qt.NoPen)
            color = self.palette().color(QPalette.Background)
            painter.setBrush(darker_or_lighter(color, 120))
            painter.drawRoundedRect(self.rect(), radius, radius)
            painter.restore()


class SelfPaintAbstractIconTextButton(SelfPaintAbstractButton):

    def __init__(self, text='', height=30, padding=0.25, parent=None):
        super().__init__(parent=parent)

        self._padding: int = int(height * padding if padding < 1 else padding)
        self._text_width = self.fontMetrics().horizontalAdvance(text)
        self._text = text

        self.setFixedHeight(height)
        self.setMinimumWidth(height + self._text_width)

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        self.paint_border_bg_when_hover(painter)
        self.draw_icon(painter)
        self.draw_text(painter)

    def draw_text(self, painter):
        text_rect = QRectF(
            self.height(), 0,
            self.width() - self.height() - self._padding, self.height()
        )
        painter.drawText(text_rect, Qt.AlignVCenter | Qt.AlignLeft, self._text)

    def draw_icon(self, painter):
        raise NotImplementedError


class SelfPaintAbstractSquareButton(SelfPaintAbstractButton):

    def __init__(self, length=30, padding=0.25, parent=None):
        """
        All buttons should has similar paddings.
        """
        super().__init__(parent)
        self._padding = int(length * padding) if padding < 1 else int(padding)
        self._body_width = length - 2 * self._padding
        self._body_rect = QRect(
            self._padding, self._padding, self._body_width, self._body_width
        )
        self.setFixedSize(length, length)

    def paint_round_bg(self, painter):
        painter.save()
        painter.setPen(Qt.NoPen)
        color = self.palette().color(QPalette.Background)
        painter.setBrush(darker_or_lighter(color, 120))
        painter.drawEllipse(self.rect())
        painter.restore()

    def paint_round_bg_when_hover(self, painter):
        opt = QStyleOptionButton()
        self.initStyleOption(opt)
        if opt.state & QStyle.State_MouseOver:
            self.paint_round_bg(painter)


class ArrowAbstractButton(SelfPaintAbstractSquareButton):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        width = self.width()
        half = width // 2
        half_half = half // 2
        self._top = QPoint(half, half_half)
        self._left = QPoint(half_half, half)
        self._right = QPoint(half_half + half, half)
        self._bottom = QPoint(half, half + half_half)

        self.vertexes = [self._top, self._left, self._right, self._bottom]

    @property
    def cross(self):
        raise NotImplementedError

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.paint_round_bg_when_hover(painter)
        set_pen_1_5(painter)
        for vertex in self.vertexes:
            if vertex != self.cross:
                painter.drawLine(self.cross, vertex)


class LeftArrowButton(ArrowAbstractButton):

    @property
    def cross(self):
        return self._left


class RightArrowButton(ArrowAbstractButton):

    @property
    def cross(self):
        return self._right


class SearchSwitchButton(SelfPaintAbstractSquareButton):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._plus_drawer = PlusIconDrawer(self.width(), self._padding)
        self._search_drawer = SearchIconDrawer(self.width(), self._padding)
        # 0.21 = 1.414 / 2 - 0.5
        self._translate_p = QPoint(self.width()//2, -int(self.width()*0.21))

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.paint_round_bg_when_hover(painter)

        # When the button size is very large, the line and the ellipse
        # will not be together.
        if self.isChecked():
            with painter_save(painter):
                painter.translate(self._translate_p)
                painter.rotate(45)
                self._plus_drawer.draw(painter)
        else:
            self._search_drawer.draw(painter)


class SettingsButton(SelfPaintAbstractSquareButton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setToolTip('配置')

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.paint_round_bg_when_hover(painter)

        pen = painter.pen()
        pen.setWidth(2)
        painter.setPen(pen)
        x = self.width() // 2
        painter.drawPoint(QPoint(x, int(self.width() * 0.3)))
        painter.drawPoint(QPoint(x, int(self.width() * 0.5)))
        painter.drawPoint(QPoint(x, int(self.width() * 0.7)))


class PlusButton(SelfPaintAbstractSquareButton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.drawer = PlusIconDrawer(self.width(), self._padding)

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.paint_round_bg_when_hover(painter)
        self.drawer.draw(painter)


class TriagleButton(SelfPaintAbstractSquareButton):

    def __init__(self, direction='up', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.drawer = TriangleIconDrawer(
            self.width(), self._padding, direction=direction
        )

    def set_direction(self, direction):
        self.drawer.set_direction(direction)

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.paint_round_bg_when_hover(painter)
        self.drawer.draw(painter)


class RecentlyPlayedButton(SelfPaintAbstractIconTextButton):

    def __init__(self, text='最近播放', **kwargs):
        super().__init__(text, **kwargs)

    def draw_icon(self, painter):
        pen_width = 1.5
        pen = painter.pen()
        pen.setWidthF(pen_width)
        painter.setPen(pen)
        x = y = self._padding
        length = self.height() - self._padding * 2
        center = self.height() // 2
        icon_rect = QRect(x, y, length, length)
        spacing = int(pen_width) + 3
        painter.drawArc(icon_rect, 250 * 16, 280 * 16)
        painter.drawLine(
            QPoint(center, center),
            QPoint(self.height() - self._padding - spacing - self._padding // 3, center)
        )
        painter.drawLine(QPoint(center, center), QPoint(center, self._padding + spacing))
        pen.setWidthF(pen_width * 2)
        painter.setPen(pen)
        painter.drawPoint(QPoint(self._padding, center))


class DiscoveryButton(SelfPaintAbstractIconTextButton):

    def __init__(self, text='发现', **kwargs):
        super().__init__(text=text, **kwargs)

        self._timer = QTimer(self)

        length = self.height()
        self._half = length // 2
        self._v1 = self._half - self._padding
        self._v2 = self._v1 / 2.5

        self._triagle = QPainterPath(QPointF(-self._v2, 0))
        self._triagle.lineTo(QPointF(0, self._v1))
        self._triagle.lineTo(QPointF(0, self._v2))
        self._rotate = 0
        self._rotate_mod = 360

        # self._timer.timeout.connect(self.on_timeout)
        # self._timer.start(30)

    def on_timeout(self):
        self._rotate = (self._rotate + 2) % self._rotate_mod
        self.update()

    def draw_icon(self, painter: QPainter):
        opt = QStyleOptionButton()
        self.initStyleOption(opt)
        set_pen_1_5(painter)
        painter.save()
        painter.translate(self._half, self._half)
        painter.rotate(self._rotate)
        for ratio in range(4):
            painter.rotate(90 * ratio)
            painter.drawPath(self._triagle)
        painter.restore()


class HomeButton(SelfPaintAbstractIconTextButton):

    def __init__(self, *args, **kwargs):
        super().__init__('主页', *args, **kwargs)
        self.home_icon = HomeIconDrawer(self.height(), self._padding)

    def draw_icon(self, painter):
        self.home_icon.paint(painter)


class CalendarButton(SelfPaintAbstractIconTextButton):

    def __init__(self, text='日历', *args, **kwargs):
        super().__init__(text, *args, **kwargs)
        self.calendar_icon = CalendarIconDrawer(self.height(), self._padding)

    def draw_icon(self, painter):
        self.calendar_icon.paint(painter)


class RankButton(SelfPaintAbstractIconTextButton):

    def __init__(self, text='排行榜', *args, **kwargs):
        super().__init__(text, *args, **kwargs)
        self.rank_icon = RankIconDrawer(self.height(), self._padding)

    def draw_icon(self, painter):
        self.rank_icon.paint(painter)


class StarButton(SelfPaintAbstractIconTextButton):

    def __init__(self, text='收藏', *args, **kwargs):
        super().__init__(text, *args, **kwargs)
        self.star_icon = StarIconDrawer(self.height(), self._padding)

    def draw_icon(self, painter):
        self.star_icon.paint(painter)


class HotButton(SelfPaintAbstractIconTextButton):
    def __init__(self, text='热门', *args, **kwargs):
        super().__init__(text, *args, **kwargs)
        self.hot_icon = FireIconDrawer(self.height(), self._padding)

    def draw_icon(self, painter):
        self.hot_icon.paint(painter)


class EmojiButton(SelfPaintAbstractIconTextButton):
    def __init__(self, emoji: str, text='表情', *args, **kwargs):
        super().__init__(text, *args, **kwargs)
        self.emoji_icon = EmojiIconDrawer(emoji, self.height(), self._padding)

    def draw_icon(self, painter):
        self.emoji_icon.paint(painter)


class PlayButton(SelfPaintAbstractSquareButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.drawer = TriangleIconDrawer(
            self.width(), self._padding, direction='right', brush=True
        )

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.paint_round_bg_when_hover(painter)
        self.drawer.draw(painter)


class PlayPauseButton(SelfPaintAbstractSquareButton):

    def __init__(self, *args, draw_circle=True, **kwargs):
        super().__init__(*args, **kwargs)

        self._draw_circle = draw_circle
        self.setCheckable(True)
        self.drawer = TriangleIconDrawer(
            self.width(), self._padding, direction='right', brush=True
        )
        self._radius = self.width() // 2

        # Calculate line rect properties.
        p0 = self.drawer.triangle[0]
        p1 = self.drawer.triangle[1]
        p2 = self.drawer.triangle[2]
        small_y = min(p1.y(), p2.y())
        big_y = max(p1.y(), p2.y())
        height = big_y - small_y
        x = p0.x()
        x2 = p1.x()
        x_move = abs(x2 - x) / 8
        self._line_half_width = 1
        self._line1_rect = QRectF(
            x - x_move - self._line_half_width, small_y, self._line_half_width * 2,
            height
        )
        self._line2_rect = QRectF(
            x2 + x_move - self._line_half_width, small_y, self._line_half_width * 2,
            height
        )
        self._translate_x = -0.125 * self._body_width
        self._pen_width = 1.5
        self._inner_rect = QRectF(
            self._pen_width, self._pen_width,
            self.width() - 2 * self._pen_width,
            self.height() - 2 * self._pen_width
        )

    def set_draw_circle(self, draw_circle: bool):
        self._draw_circle = draw_circle
        self.update()

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        set_pen_width(painter, self._pen_width)
        self.paint_round_bg_when_hover(painter)
        if self.isChecked():
            with painter_save(painter):
                painter.translate(self._translate_x, 0)
                painter.setBrush(painter.pen().color())
                painter.drawRoundedRect(
                    self._line1_rect, self._line_half_width, self._line_half_width
                )
                painter.drawRoundedRect(
                    self._line2_rect, self._line_half_width, self._line_half_width
                )
        else:
            self.drawer.draw(painter)
        if self._draw_circle is True:
            painter.drawEllipse(self._inner_rect)


class _PlayXButton(SelfPaintAbstractSquareButton):

    def __init__(self, direction, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.drawer = TriangleIconDrawer(
            self.width(), self._padding, direction=direction, brush=False
        )
        p0 = self.drawer.triangle[0]
        p1 = self.drawer.triangle[1]
        p2 = self.drawer.triangle[2]

        # When the directory is 'right'
        # left distance (before): body_width // 4
        # left distance (now): body_width//2 - body_width*0.75//2
        self._body_translate = -0.125 if direction == 'right' else 0.125

        # Line rect properties.
        small_y = min(p1.y(), p2.y())
        big_y = max(p1.y(), p2.y())
        half_width = 1.5
        height = big_y - small_y
        x = p0.x()
        self._line = QPointF(x, small_y), QPointF(x, big_y)
        self._line_rect = QRectF(x - half_width, small_y, half_width * 2, height)

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.paint_round_bg_when_hover(painter)
        with painter_save(painter):
            painter.translate(self._body_translate * self._body_width, 0)
            self.drawer.draw(painter)
            with painter_save(painter):
                set_pen_1_5(painter)
                painter.setBrush(painter.pen().color())
                painter.drawLine(*self._line)


class PlayNextButton(_PlayXButton):

    def __init__(self, *args, **kwargs):
        super().__init__('right', *args, **kwargs)


class PlayPreviousButton(_PlayXButton):

    def __init__(self, *args, **kwargs):
        super().__init__('left', *args, **kwargs)


class MVButton(SelfPaintAbstractSquareButton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.rectangle_height = int(self.width() * 0.618)
        self.rectangle_y = (self.height() - self.rectangle_height) // 2
        self.drawer = TriangleIconDrawer(
            self.width(), self.width() * 0.35, direction='right', brush=True
        )

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        opt = QStyleOptionButton()
        self.initStyleOption(opt)
        rect = QRect(1, self.rectangle_y, self.width() - 2, self.rectangle_height)
        with painter_save(painter):
            set_pen_1_5(painter)
            if opt.state & QStyle.State_MouseOver:
                color = self.palette().color(QPalette.Background)
                painter.setBrush(darker_or_lighter(color, 120))
            painter.drawRoundedRect(rect, 3, 3)
        self.drawer.draw(painter)


class VolumeButton(SelfPaintAbstractSquareButton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.drawer = VolumeIconDrawer(self.width(), self._padding)

    def set_volume(self, volume):
        self.drawer.set_volume(volume)
        self.update()

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        self.paint_round_bg_when_hover(painter)
        self.drawer.draw(painter, self.palette())


if __name__ == '__main__':
    from PyQt5.QtWidgets import QHBoxLayout, QVBoxLayout

    from feeluown.gui.debug import simple_layout

    length = 40

    with simple_layout(QVBoxLayout) as layout:
        l1 = QHBoxLayout()
        l2 = QHBoxLayout()
        l3 = QVBoxLayout()
        layout.addLayout(l1)
        layout.addLayout(l2)
        layout.addLayout(l3)

        l1.addWidget(LeftArrowButton(length=length))
        right = RightArrowButton(length=length)
        right.setDisabled(True)
        l1.addWidget(right)
        l1.addWidget(SearchSwitchButton(length=length))
        l1.addWidget(SettingsButton(length=length))
        l1.addWidget(RecentlyPlayedButton(height=length))

        l1.addWidget(TriagleButton(length=length, direction='up'))
        l1.addWidget(CalendarButton(height=length))

        l2.addWidget(PlayPreviousButton(length=length))
        l2.addWidget(PlayPauseButton(length=60))
        l2.addWidget(PlayNextButton(length=length))
        volume_button = VolumeButton(length=length)
        volume_button.set_volume(60)
        l2.addWidget(volume_button)
        l2.addStretch(0)

        l3.addWidget(HotButton(height=length))
        l3.addWidget(HomeButton(height=length))
        l3.addWidget(DiscoveryButton(height=length))
        l3.addWidget(RankButton(height=length))
        l3.addWidget(StarButton(height=length))
        l3.addWidget(EmojiButton('😁', '开心', height=length))
        l3.addWidget(EmojiButton('🔥', '热门', height=length))
        l3.addStretch(0)
